*+
*+
*+    Source Module => OPENFILE.PRG
*+
*+    Functions: Function SA_DBUSEAREA()
*+               Static Function SA_DBUSE1()
*+               Static Function UseErr()
*+               Function doserr()
*+               Function ErrTxt()
*+
*+       Tables: dbusearea( plNewArea, pcDriver, pcName, pxcAlias, plShared, mlReadOnly )
*+               dbusearea( plNewArea, pcDriver, pcName, pxcAlias, plShared, mlReadOnly )
*+
*+    Reformatted by Click! 1.13 on Jan-31-1998 at  1:30 pm
*+
*+

/*
Problem: An application using several .DBF-files in a set of directories
on a server. Each user or group of users has different sets of access
rights to these directories.

Solution: My special UDF that opens the file in a mode according to each
users effective rights: If he lacks Write-access the file is opened in
ReadOnly mode. If he does not have the required access rights, he is
told so. All without the application bombing him with error messages.

The reason for why I post it here is to hear others opinions on my
solution: Is this a proper way of handling this kind of problem? Do you
see any pitfalls in my source code? (yes, I have of course tested it).
Do you have a better solution?

If you need it, use it yourself. But please make public any major
improvements or error corrections.
*/

//* DBUSE.PRG
//
// Stian M. Abrahamsen 03.12.97
// Email: stiana@online.no

#include "fileio.ch"
#include "error.ch"

#define READONLY    .T.
#define SHARED      .T.
#define CORRUPTION   1
#define TRY_RO_MODE  2
#define TRYAGAIN     3

*+
*+
*+    Function SA_DBUSEAREA()
*+
*+    Called from ( openfile.prg )   1 - function sa_dbusearea()
*+
*+
*+
function SA_DBUSEAREA( pcName, pcOpenMode, plShared, plNewArea, pxcAlias, pcDriver )

/*
FUNCTION SA_DBUSEAREA([<pcName>,[<pcOpenMode>],[<plShared>],[<plNewArea>],[<pxcAlias>],[<pcDriver>])

Parameters:
pcName:     The database file to be used.
pcOpenMode: 'RO': open the database file for read-only access.
            'RW': permit updates to the database file.
            'PRW': "preferably read/Write". If possible, the database file is
            opened for read/write access. If not (i.e. not sufficient access
            rights, or the read-only file attribute is set) it is opened for
            read-only access.
            Defaults to 'PRW'
            NB!: Only access limitations and not concurrency issues should lead
            to the file being opened read-only!
plShared:   .T.: allow access for others. .F.: claim exclusive access for this
            process. Defaults to the value set by _SET_EXCLUSIVE.
plNewArea:  .F.: use current work area. .T.: use lowest numbered unoccupied work
            area. Defaults to .F.
pxcAlias:   If specified, alias for the work area to be used instead of pcName
pcDriver:   If not specified, the default database driver is used. (DBFNTX, or
            the most recent call to DBSETDRIVER())

Returns: 'RO': successfully opened for read-only access.
         'RW': successfully opened for read/write.
         'C':  could not be opned due to concurrency issues.
         '':   Could not be opened due to other errors.
*/

local mcType      := ''
local mlNoError   := .T.
local mnCount     := 5
local mbPrevErr
local mlReadOnly  := .T.
local mnBreakCode := 0
local mcResult    := ''

//Set defaults
if ( mcType := valtype( pcName ) ) <> 'C'
   if mcType == 'U'
      //ERROR! No filename
      alert( 'Missing filename in SA_DBUSEAREA()' )
   else
      //ERROR! Filename must be of type character
      alert( 'Filename must be character type in SA_DBUSEAREA()' )
   endif
endif

if ( mcType := valtype( pcOpenMode ) ) <> 'C'
   if mcType == 'U'
      //Set default value
      pcOpenMode := 'PRW'
   else
      //Parameter error
      alert( 'Parameter error for OpenMode in SA_DBUSEAREA()' )
   endif
else
   //Parameter value within the set of legal values?
   if !ascan( { 'RO', 'RW', 'PRW' }, { | x | upper( pcOpenMode ) == x } ) > 0
      alert( 'Parameter error for OpenMode on SA_DBUSEAREA()' )
   endif
   pcOpenMode := upper( pcOpenMode )
endif

if ( mcType := valtype( plShared ) ) <> 'L'
   if mcType == 'U'
      //Set default value
      plShared := .T.
   else
      //Parameter error
      alert( 'Parameter error for Shared in SA_DBUSEAREA()' )
   endif
endif

if ( mcType := valtype( plNewArea ) ) <> 'L'
   if mcType == 'U'
      //Set default value
      plNewArea := .F.
   else
      //Parameter error
      alert( 'Parameter error for NewArea in SA_DBUSEAREA()' )
   endif
endif

if ( mcType := valtype( pxcAlias ) ) <> 'C'
   if mcType == 'U'
      //Default value not required
   else
      //Parameter error
      alert( 'Paramater error for Alias in SA_DBUSEAREA()' )
   endif
endif

if ( mcType := valtype( pcDriver ) ) <> 'C'
   if mcType == 'U'
      //Set default value
      pcDriver := dbsetdriver()
   else
      //Parameter error
      alert( 'Parameter error for databasedriver in SA_DBUSEAREA()' )
   endif
endif

mlReadOnly := if( pcOpenMode == 'RO', .T., .F. )
mcResult   := if( mlReadOnly == .T., 'RO', 'RW' )

mbPrevErr := errorblock( { | e | UseErr( e, @mlNoError, @mcResult, pcOpenMode, @mlReadOnly, @mnBreakCode, mbPrevErr ) } )

begin sequence
   dbusearea( plNewArea, pcDriver, pcName, pxcAlias, plShared, mlReadOnly )
recover
   do case
   case mnBreakCode == CORRUPTION
      //File corruption shold be handled here
      alert( 'Corruption in database file ' + pcName + '.;Program Aborts.' )
      mcResult := ''
      quit
   case mnBreakCode == TRY_RO_MODE
      //Recursively calling SA_DBUSEAREA()
      if SA_DBUSEAREA( pcName, 'RO', plShared, plNewArea, pxcAlias, pcDriver ) == 'RO'
         //Success!
         mcResult := 'RO'
      else
         //Could not open, even in RO-mode
         mlNoError := .F.
         mcResult  := ''
      endif
   case mnBreakCode == TRYAGAIN
      //Another user has opened the file. Try over again for a few times.
      do while !( mlNoError := SA_DBUSE1( pcName, pcOpenMode, plShared, plNewArea, pxcAlias, pcDriver ) ) .and. mnCount -- > 0
         inkey( 0.1 )                   //A tiny delay
      enddo
      if mlNoError
         //Success!
         mcResult := if( mlReadOnly == .T., 'RO', 'RW' )
      else
         mcResult := 'C'
      endif
   endcase
end

errorblock( mbPrevErr )

return mcResult

*+
*+
*+    Static Function SA_DBUSE1()
*+
*+    Called from ( openfile.prg )   1 - function sa_dbusearea()
*+
*+
*+
static function SA_DBUSE1( pcName, pcOpenMode, plShared, plNewArea, pxcAlias, pcDriver )

local mlNoError   := .T.
local mbPrevErr
local mlReadOnly  := if( pcOpenMode == 'RO', .T., .F. )
local mnBreakCode := 0
local mcResult    := if( mlReadOnly == .T., 'RO', 'RW' )

mbPrevErr := errorblock( { | e | UseErr( e, @mlNoError, @mcResult, pcOpenMode, @mlReadOnly, @mnBreakCode, mbPrevErr ) } )

begin sequence
   dbusearea( plNewArea, pcDriver, pcName, pxcAlias, plShared, mlReadOnly )
recover
   //Another user has opened the file.
end

errorblock( mbPrevErr )

return mlNoError

*+
*+
*+    Static Function UseErr()
*+
*+    Called from ( openfile.prg )   1 - function sa_dbusearea()
*+                                   1 - static function sa_dbuse1()
*+
*+
*+
static function UseErr( oErr, plNoError, pcResult, pcOpenMode, plReadOnly, pnBreakCode, pbMyAncestor )

//OS- and file system related errors:
if oErr:genCode == EG_OPEN .and. oErr:osCode == 53
   //Server not found (UNC)
   plNoError := .F.
   pcResult  := ''
   neterr( .T. )
   return .F.
endif
if oErr:genCode == EG_OPEN .and. oErr:osCode == 67
   //Volume/share not found (UNC)
   plNoError := .F.
   pcResult  := ''
   neterr( .T. )
   return .F.
endif
if oErr:genCode == EG_OPEN .and. oErr:osCode == 3
   //Path not found
   plNoError := .F.
   pcResult  := ''
   neterr( .T. )
   return .F.
endif
if oErr:genCode == EG_OPEN .and. oErr:osCode == 2
   //File not found
   plNoError := .F.
   pcResult  := ''
   neterr( .T. )
   return .F.
endif
if oErr:genCode == EG_OPEN .and. oErr:osCode == 4
   //Too many open files
   plNoError := .F.
   pcResult  := ''
   neterr( .T. )
   return .F.
endif
if oErr:genCode == EG_DUPALIAS .and. oErr:osCode == 0
   //Alias already in use
   //NB! Uses a file handle!
   plNoError := .F.
   pcResult  := ''
   neterr( .T. )
   return .F.
endif
if oErr:genCode == EG_CORRUPTION .and. oErr:osCode == 0
   //File corruption
   plNoError   := .F.
   pcResult    := ''
   pnBreakCode := CORRUPTION
   neterr( .T. )
   break
endif

//Erors related to network access rights and the RO file attribute:
//File exists, but user don't have proper access rights:
//  Tries to open ReadOnly but lacks Read-access
//  Tries to open Read/Write but the RO file attribute is set
//  Tries to open Read/Write but lacks Read- and Write-access
if oErr:genCode == EG_OPEN .and. oErr:osCode == 5
   if pcOpenMode == 'PRW' .and. plReadOnly == .F.
      //User tried to open in Read/Write mode and pcOpenMode is 'PRW'
      //Try over again, but in ReadOnly mode
      plReadOnly  := .T.
      pnBreakCode := TRY_RO_MODE
      break
   endif
   plNoError := .F.
   pcResult  := ''
   neterr( .T. )
   return .F.
endif

//Concurrency related erors:
//When another user has opened the file and:
//   user tries to open in Exclusive mode
//   or the other user has opened the file in exclusive mode
//In this case, retry opening the file
if oErr:genCode == EG_OPEN .and. oErr:osCode == 32
   plNoError := .F.
   pcResult  := 'C'
   neterr( .T. )
   pnBreakCode := TRYAGAIN
   break
endif

//If other kinds of errorrs occur let the previous error handler take
//care of them.
alert( ErrTxt( oErr ) )                 //Report the attributes of the error object

return eval( pbMyAncestor, oErr )

*+
*+
*+    Function doserr()
*+
*+
*+
function doserr( pcDosErrNr )

local DosErr := { "Invalid function number", ;
                  "File not found", ;
                  "Path not found", ;
                  "Too many open files", ;
                  "Access denied", ;
                  "Invalid handle", ;
                  "Memory", ;
                  "Insufficient memory", ;
                  "Memory", ;
                  "Invalid environment", ;
                  "Invalid format", ;
                  "Invalid access code", ;
                  "Invalid data", ;
                  "", ;
                  "Invalid drive", ;
                  "Cannot remove directory", ;
                  "Not same device", ;
                  "No more files", ;
                  "Write-protect", ;
                  "Unknown unit", ;
                  "Drive not ready", ;
                  "Unknown command", ;
                  "Data error (CRC)", ;
                  "Bad request", ;
                  "Seek error", ;
                  "Unknown media type", ;
                  "Sector not found", ;
                  "Printer out of paper", ;
                  "Write fault", ;
                  "Read fault", ;
                  "General failure", ;
                  "Sharing violation", ;
                  "Lock violation", ;
                  "Invalid disk change", ;
                  "FCB unavailable", ;
                  "Sharing buffer overflow", ;
                  "", ;
                  "", ;
                  "", ;
                  "", ;
                  "", ;
                  "", ;
                  "", ;
                  "", ;
                  "", ;
                  "", ;
                  "", ;
                  "", ;
                  "", ;
                  "Network request not supported", ;
                  "Remote not listening", ;
                  "Duplicate name on network", ;
                  "Network name not found", ;
                  "Network busy", ;
                  "Network device no longer exists", ;
                  "Network BIOS command limit exceeded", ;
                  "Network adapter hardware error", ;
                  "Incorrect response from network", ;
                  "Unexpected network error", ;
                  "Incompatible remote adapter", ;
                  "Print queue full", ;
                  "Not enough space for print file", ;
                  "Print file deleted", ;
                  "Network name deleted", ;
                  "Access denied", ;
                  "Network device type incorrect", ;
                  "Network name not found", ;
                  "Network name limit exceeded", ;
                  "Network BIOS session limit exceeded", ;
                  "Temporarily paused", ;
                  "Network request not accepted", ;
                  "Print or disk redirection paused", ;
                  "", ;
                  "", ;
                  "", ;
                  "", ;
                  "", ;
                  "", ;
                  "", ;
                  "File already exists", ;
                  "", ;
                  "Cannot make directory entry", ;
                  "Fail on INT 24H", ;
                  "Too many redirections", ;
                  "Duplicate redirection", ;
                  "Invalid password", ;
                  "Invalid parameter", ;
                  "Network device fault" }

return DosErr[ pcDosErrNr ]

*+
*+
*+    Function ErrTxt()
*+
*+    Called from ( openfile.prg )   1 - static function useerr()
*+
*+
*+
function ErrTxt( oErr )

local mcMelding := ''
local i         := 0

mcMelding := 'Error Information:' + ';'

if valtype( oErr:args ) == 'A'
   mcMelding += ";LEN(oErr:args)     : " + ltrim( str( len( oErr:args ) ) )
   for i := 1 to len( oErr:args )
      mcMelding += ';oErr:args[' + ltrim( str( i ) ) + '] type: ' + valtype( oErr:args[ i ] )
   next
endif

mcMelding += ;
        ";canDefault    : " + if( oErr:canDefault, '.T.', '.F.' ) ;
        + ";canRetry      : " + if( oErr:canRetry, '.T.', '.F.' ) ;
        + ";canSubstitute : " + if( oErr:canSubstitute, '.T.', '.F.' ) ;
        + ";VALTYPE(cargo): " + valtype( oErr:cargo ) ;
        + ";description   : " + oErr:description ;
        + ";filename      : " + oErr:filename ;
        + ";genCode       : " + ltrim( str( oErr:genCode ) ) ;
        + ";operation     : " + oErr:operation

if !oErr:oscode == 0
   mcMelding += ";osCode " + ltrim( str( oErr:osCode ) ) + ': ' + doserr( oErr:osCode )
endif

mcMelding += ;
        + ";severity      : " + ltrim( str( oErr:severity ) ) ;
        + ";subCode       : " + ltrim( str( oErr:subCode ) ) ;
        + ";subSystem     : " + oErr:subSystem ;
        + ";tries         : " + ltrim( str( oErr:tries ) )

i := 2
do while !procname( i ) == ""
   mcMelding += ";Kalt fra: " + procname( i ) + ' (' + ltrim( str( procline( i ) ) ) + ')'
   i ++
enddo

return mcMelding

*+ EOF: OPENFILE.PRG
